! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: mosspectra_mod
!
! Authors:  Brendan Perry     - NASA/GSFC, brendan.h.perry@nasa.gov 
!           Kip Kuntz         - NASA/GSFC/JHU kuntz@pha.jhu.edu
!           Steven L. Snowden - NASA/GSFC (original algorithm)
! 
! Description: Former Extended Source Analysis Software subtask mos-spectra
!
! Comments:
!
! $Id: mosspectra_mod.f90,v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!  Version 0.11 - complete rewrite of sas task mos-spectra perl into f90 by BP
!
! This task creates intermediary files for creation of model particle background
! spectra and images by mosback. This is a functionally equivalent but complete 
! f95 rewrite of the original perl mos-spectra
!
!  usage: mosspectra + 
!  eventfile -- event list from mosfilter or espfilt
!  cornerfile -- corner event list from mosfilter or espfilt
!  imagefile -- Entire FOV region observation image.
!  expmap -- Exposure map in DET coords
!  spmask -- Mask for calculating soft proton scaling (FOV image for all events)
!  rmffile -- Response matrix file name (will construct name if none given).
!  arffile -- Auxillary response matrix file name (will construct name if none given).
!  keepinterfiles -- Keep intermediate files created (F to delete)
!  withregion -- Input a text region file?
!  regionfile -- Text file with additional region information.
!  withsrcrem -- Perform source exclusion (with source removal)?
!  masksky SKY -- SKY coord mask FITS file name.
!  maskdet DET -- DET coord mask FITS file name.
!  elow -- Band low energy limit (elow).
!  ehigh -- Band high energy limit (ehigh).
!  ccds -- MOS CCD (1-7) T/F to include/ignore
!  pattern -- evselect PATTERN value (usually 12)
!
!  The event file should be cleaned of bad time intervals,
!  bad pixels, and the like, but MUST include the corner pixels.
!  (So PATTERN<=12 is good, but FLAG==0 is not!)
!
!  OLD: SAS must be set up and running in the window
!  obj - observation data from the selected region
!  oc - observation data from the corners
!  ff - filter wheel closed from the selected region
!  fc - filter wheel closed from the corners
!
module mosspectra_mod

  use dal
  use cal
  use param
  use dataFilter
  use errorHandling
  use sasdal
  use gofio_mod
  use gofutils_mod

  implicit none

  private

  character(len=*), parameter :: &
       eventpar    = "eventfile", &       ! input EMOS event list from espfilt
       cornerpar   = "cornerfile", &      ! input EMOS event list from espfilt
       imagepar    = "imagefile", &       ! input EMOS image from espfilt
       expmappar   = "expmap", &          ! input EMOS exposure map
       spmappar    = "spmask", &          ! input EMOS SP map
       specpar     = "specfile", &        ! input EMOS FOV spectrum
       maskpar     = "mask", &            ! input EMOS mask 
       withregpar  = "withregion", &      ! Apply region file?
       regpar      = "regionfile", &      ! text region file
       withsrcXpar = "withsrcrem", &      ! Apply background region in sky coords?
       skypar      = "masksky", &         ! input EMOS bkg region file
       detpar      = "maskdet", &         ! input EMOS bkg region file
       rmfpar      = "rmffile", &         ! response matrix file
       arfpar      = "arffile",&          ! auxilliary response file
       ccdspar     = "ccds", &            ! user-selected ccds
       elowpar     = "elow", &            ! Low energy limit for the bands (eV)
       ehighpar    = "ehigh", &           ! High energy limit for the bands (eV)
       pattpar     = "pattern",&          ! evselect PATTERN
       keeppar     = "keepinterfiles",&   ! keep (delete) intermediary files?
       pixrespar   = "badpixelresolution" ! add param to calls of backscale
       
  public mosspectra

contains

  !---------------------------------------------------------------------
  subroutine mosspectra
    !---------------------------------------------------------------------

    type(DataSetT) :: inEVLI_S, inSKY_S, inDET_S, temp_S, FWC_S, &
         ccdIMG_S, ccdSPEC_S, fwcIMG_S, inSPEC_S
    type(ArrayT) :: inEVLI_A
    type(TableT) :: inEVLI_T, FWC_T
    type(ColumnT) :: expo_C, nflag_C, ccdnr_C
    type(CalInfo) :: fwcFileInfo
    character(len=1) :: schip, cband
    character(len=2) :: instE, cstep
    character(len=5) :: expidE, suff, inst, pattern
    character(len=6) :: celow, cehigh, filterE
    character(len=10) :: obsidE, cpixRes
    character(len=18) :: prefix, type
    character(len=25) :: smodeE, datamode
    character(len=20) :: dateobs, dateend
    character(len=60) :: maskitdet, maskitsky
    character(len=125) :: expression
    character(len=256) :: inEVLI_F, inCORN_F, inIMG_F, inEXPMAP_F, inSPMASK_F, &
         inREG_F, inSKY_F, inDET_F, inSPEC_F, inRMF_F, inARF_F, inMASK_F, inFWC_F, &
         inEXPMASK_F, inSPMAP_F, inBAND_F, ccdIMG_F, ccdSPEC_F, ccdMASK_F, ccdEXPIMG_F,&
         fwcIMG_F, tempout_F
    character(len=256) :: atthk_F, ccfdir, ccdnoise, testdir
    character(len=4096) :: command
    character(len=1024) :: corndef, fulldef, fovdef, ccddef, ccddefa
    character(len=50) :: chipdef(7)
    logical :: i_ok, warndie, die, mos1, mos2, pn, keep, go, ccds(7)
    logical :: withsrcX, withreg
    integer(kind=int8), dimension(:), pointer :: nflag_BP, ccdnr_BP
    integer :: frmtimeE, cnts, newccds(7)
    integer :: typeF, nRowsE, revolutE, ff, cc, step
    integer :: elow, ehigh, ehighE, elowE, ilo, ihi
    integer :: i, j, k, n, nEVLIs, patt, n1, n2, time
    integer :: lun, status = 0
    real :: xfave, delxf, hard, hardsig, livetime, rapnt, decpnt
    real :: hige, lowe, rate, pixRes
    real :: backE, areaE, expoE, expo, expo_CCD(7)
    real(kind=single), dimension(:), pointer :: expo_RP
    !
    inSKY_F    = ""
    inDET_F    = ""
    inREG_F    = ""
    inSPMASK_F = ""
    inMASK_F   = ""
    inSPEC_F   = ""
    inEXPMAP_F = ""
    maskitsky  = ""
    maskitdet  = ""
    !
    ! Step 1 get some parameters, open and read eventlist for keywords
    ! was 'prefix', now eventlist filename.
    ! eventlist comes from espfilt
    ! OLDNEW: mos1S001-clean.fits mos1S001-allevc.fits from espfilt
    !
    step = 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': get attributes from allevc file.'
    call message(AppMsg,Always) ! change to Verbose???!!!
    inEVLI_F = stringParameter(eventpar)  ! mos1S001-allevc.fits
    status   = 0
    warndie  = .TRUE.
    call check_open_set(trim(inEVLI_F),inEVLI_S,1,status,warndie)
    !
    ! BHP: Get critical header keywords from input eventlist (from espfilt)
    !
    status = 0
    warndie = .TRUE. ! if not found, unreadable, die
    typeF = 1 ! 1 = EVLI, 2 = SPEC, 3 = IMG, most of these are dummies
    call read_header_info(typeF,inEVLI_S,inEVLI_T,inEVLI_A,instE,obsidE,&
         expidE,smodeE,frmtimeE,filterE,revolutE,backE,areaE,expoE,elowE,&
         ehighE,nRowsE,warndie)
    write(errstr,*) "OBSID = ",obsidE," EXPID = ",expidE
    call message(AppMsg,Always)
    call release(inEVLI_S) ! BHP: don't need it open anymore
    !
    ! SS: If revolution < 42 then quit as we do not have the right files
    !     for reducing those data
    if (revolutE .lt. 42) then
       call error ("Pre42","Observation was before Rev 42. Cannot analyze.")
    endif
    !
    do i=1,7
       ccds(i) = booleanParameter(ccdspar,i-1) ! T T T T F T T
    enddo
    if (ccds(1)) then 
       if (index(smodeE,"Fast") .ne. 0) then ! check Un vs not???!!!
          call warning('FastSubWithCCD1',&
               "Fast SUBMODE incompatable with ccd1=T. Resetting ccds(1)=F, continuing. ")
          ccds(1) = .FALSE.
       endif
    endif
    !
    ! set 'prefix' to create other filenames if not specifically entered
    ! from parameters.
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': set prefix, call CAL_setState'
    call message(AppMsg,Always) ! change to Verbose???!!!
    inst = "EMOS1"
    mos1 = .TRUE.
    if (instE .eq. 'M1') then
       ! Step 2a: remove 'bad' (dead) ccds based on orbit and instrument
       call check_orbit(ccds,revolutE) ! goflib/gofio
       prefix = "mos1" // trim(adjustl(expidE)) ! mos1S001
       call CAL_setState(instrument=EMOS1)
    elseif (instE .eq. 'M2') then
       mos1 = .FALSE.
       inst = "EMOS2"
       prefix = "mos2" // trim(adjustl(expidE)) ! mos2S002
       call CAL_setState(instrument=EMOS2)
    else
       call error("notMOS","Input event list INSTRUME must be EMOS1 or EMOS2")
    endif
    ! get the name of the current EMOSn_FWC_nnnn.CCF file
    call CAL_fwcInfo(fwcFileInfo)
    !    
    inFWC_F = fwcFileInfo%fileName
    write(errstr,*) 'FWC CCF filename : ', trim(adjustl(inFWC_F))
    call message(AppMsg,Always) ! change to Verbose???!!!
    !
    ! now get the rest of the input parameters
    ! this cleaned corner event list comes from espfilt
    ! OLDNEW: mos1S001-corn.fits mos1S001-corevc.fits
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*)'Step ' // trim(adjustl(cstep)) // ': get the rest of the input parameters'
    call message(AppMsg,Always) ! change to Verbose???!!!
    inCORN_F   = stringParameter(cornerpar) ! mos1S001-corevc.fits
    if (trim(adjustl(inCORN_F)) .eq. 'default') then
       inCORN_F = trim(adjustl(prefix)) // "-corevc.fits"
    endif
    ! this cleaned fov image comes from mosspectra, cheese, or espfilt !!!
    ! OLDNEW: mos1S001-obj-im.fits/mos1S001-fovimt.fits
    inIMG_F    = stringParameter(imagepar)  ! mos1S001-fovimt.fits
    if (trim(adjustl(inIMG_F)) .eq. 'default') then
       ! fovim or fovimc or fovimt/s/ t and c should be ~same, no?
       inIMG_F = trim(adjustl(prefix)) // "-fovimt.fits"
    endif
    ! this SP scaling file is created only by mosspectra
    inSPMASK_F = stringParameter(spmappar) ! mos1S001-fovimspdet.fits
    if (trim(adjustl(inSPMASK_F)) .eq. 'default') then
       ! fovspdet t? h? s???
       inSPMASK_F = trim(adjustl(prefix)) // "-fovimspdet.fits"
    endif
    ! this mask band region file is created only by mosspectra
    inMASK_F = stringParameter(maskpar) ! mos1S001-maskimt.fits
    if (trim(adjustl(inMASK_F)) .eq. 'default') then
       ! t? h? s???
       inMASK_F = trim(adjustl(prefix)) // "-maskimt.fits"
    endif
    ! this spectrum is also created by cheese???!!!
    inSPEC_F = stringParameter(specpar) ! mos1S001-fovt.pi
    if (trim(adjustl(inSPEC_F)) .eq. 'default') then
       ! t? h? s???
       inSPEC_F = trim(adjustl(prefix)) // "-fovt.pi"
    endif
    ! this exposure map is created cheese or mosspectra
    inEXPMAP_F = stringParameter(expmappar) ! mos1S001-expimt.fits
    if (trim(adjustl(inEXPMAP_F)) .eq. 'default') then
       !  t? h? s??? can use fovimtexp file???
       inEXPMAP_F = trim(adjustl(prefix)) // "-expimt.fits"
    endif
    ! these files comes from cheese (apply a sky/det coord region mask)
    withsrcX    = booleanParameter(withsrcXpar)
    if (withsrcX) then
       inSKY_F = stringParameter(skypar)   ! mos1S001-bkgregtsky.fits, from CHEESE
       ! can trigger withsky but if no name given, creates a default name
       ! "bkg region, total band, sky coords" = bkgregtsky
       if (trim(adjustl(inSKY_F)) .eq. 'default') then
          inSKY_F = trim(adjustl(prefix)) // "-bkgregtsky.fits"
       endif
       inDET_F = stringParameter(detpar)   ! mos1S001-bkgregtdet.fits
       ! can trigger withdet but if no name given, creates a default name
       ! "bkg region, total band, det coords" = bkgregtdet
       if (trim(adjustl(inDET_F)) .eq. 'default') then
          inDET_F = trim(adjustl(prefix)) // "-bkgregtdet.fits"
       endif
    end if
    !
    ! plain text, user-selected, region file
    !
    withreg    = booleanParameter(withregpar)
    if (withreg) then
       inREG_F = stringParameter(regpar)   ! reg.txt
    end if
    ! add to backscale
    pixRes = realParameter(pixrespar)
    write(cpixRes,'(F10.3)')pixRes
    !
    ! get these things from the FWC file
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*)'Step ' // trim(adjustl(cstep)) // ': get keywords from the FWC CCF file'
    call message(AppMsg,Always) ! change to Verbose???!!!
    FWC_S = dataSet(trim(adjustl(inFWC_F)), Read)
    FWC_T = table(FWC_S,'EXPTIME')
    expo_C  = column(FWC_T,'EXPOSURE',Read)
    expo_RP => real32Data(expo_C)
    ccdnr_C  = column(FWC_T,'CCDNR',Read)
    ccdnr_BP => int8Data(ccdnr_C)
    nflag_C  = column(FWC_T,'NFLAG',Read)
    nflag_BP => int8Data(nflag_C)
    expo_CCD = 0.
    do i=1,7
       ! get exposures for each ccd from FWC data
       if (ccds(i)) then
          write(cband,'(I1)')i
          expo_CCD(i) = sum(expo_RP,MASK=((nflag_BP.eq.0).and.(ccdnr_BP.eq.i)))
          write(errstr,*)&
               'FWC exposure for selected CCD'// cband // ' = ',expo_CCD(i)
          call message(AppMsg,Always) ! change to Verbose???!!!
       end if
    enddo
    call release(FWC_S)
    ! user-specified, but if still 'default', generate rmf as usual 
    inRMF_F = stringParameter(rmfpar)
    inARF_F = stringParameter(arfpar)
    if (trim(adjustl(inRMF_F)) .eq. 'default') then
       inRMF_F = trim(adjustl(prefix)) // ".rmf"
    end if
    if (trim(adjustl(inARF_F)) .eq. 'default') then
       inARF_F = trim(adjustl(prefix)) // ".arf"
    endif
    !
    keep   = booleanParameter(keeppar)        ! T or F, delete intermediary files
    !
    celow = ''
    cehigh = ''
    elow  = intParameter(elowpar)  ! 0-11999, constrained in params, check these???!!!
    ehigh = intParameter(ehighpar) ! 1-12000, constrained in params
    write(celow,'(I5)') elow
    write(cehigh,'(I5)') ehigh
    !
    patt = intParameter(pattpar)  ! evselect PATTERN
    write(pattern,'(I2)') patt
    pattern = trim(adjustl(pattern)) // ")"
    !
    ! BHP: added withimages 10/2021. Replaces 'if (ehigh >0)' schema. If
    ! false, only spectra will be created.
    ! BHP: withimages removed 01/2023 (confuses downstream esas)
    !
    call printParams(inEVLI_F,inCORN_F,inIMG_F,inSPMASK_F,inEXPMAP_F,inMASK_F,&
         withsrcX,inSKY_F,inDET_F,withreg,inREG_F,inRMF_F,inARF_F,ccds,&
         elow,ehigh,keep,patt)
    !
    ! Step 5: Definitions
    !   corndef is a definition of the corner pixels
    !   chipdef is a definition of the location of each chip
    !   fulldef is a definition of the full FOV in detector coordinates
    !   ccddef is the selection criteria for the active CCDs, i.e., 
    !        those not excluded by the ccds parameter
    ! checks for CCD losses removed for SAS-19 (handled by reprocessed ODFs)
    !
    ! Set the CCD region selection.  This gets a bit convoluted as
    ! eexpmap doesn't exclude regions but does exclude CCDs while 
    ! arfgen calculates BACKSCAL with regions but not with CCD selection.
    !
    ccddef  = ""
    ccddefa = ""
    corndef = ""
    chipdef = ""
    fulldef = ""
    !    
    call defineCCDregion (instE,ccds,ccddef,ccddefa)
    call defineexpr (inst,corndef,chipdef,fulldef)
    ! set these to Verbose???!!!
    step = step + 1
    write(cstep,'(I2)')step
    if (ccddef .eq. "") then
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: ccddef = N/A All CCDs selected'
    else
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: ccddef = ',trim(ccddef)
    endif
    call message(AppMsg,Always)
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: ccddefa = ',trim(ccddefa)
    call message(AppMsg,Always)
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: corndef = ',trim(corndef)
    call message(AppMsg,Always)
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: fulldef = ',trim(fulldef)
    call message(AppMsg,Always)
    !
    ! Step 4  Add the source exclusion file if desired
    ! NOTE: need BOTH
    !
    maskitsky = ""
    maskitdet = ""
    if (withsrcX) then ! if withsrcX and OLDNEW: mos-bkg_region-sky.fits mos1S001-bkgregtsky.fits
       step = step + 1
       write(cstep,'(I2)')step
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Adding the source exclusion files'
       call message(AppMsg,Always)
       status = 0 !                      
       warndie = .FALSE.
       call check_open_set(trim(inSKY_F),inSKY_S,1,status,warndie)
       if (status .eq. 1) then
          write(errstr,*) "SKY region file entered: ",trim(adjustl(inSKY_F)),", exists."
          call message(AppMsg,Always)
          maskitsky = "&&region(" // trim(adjustl(inSKY_F)) // ")"
          status = 0
          warndie = .FALSE.
          call check_open_set(trim(inDET_F),inDET_S,1,status,warndie)
          if (status .eq. 1) then
             write(errstr,*) &
                  "DET region file entered also exists, will proceed with source exclusion."
             call message(AppMsg,Always)
             maskitdet = "&&region(" // trim(adjustl(inDET_F)) // ")"
          else
             ! SS: Check to see if the files exists
             write(errstr,*) &
                  "DET region file does NOT exist. Will NOT proceed with source exclusion."
             call message(AppMsg,Always)
             maskitsky = "" ! need both, so if no DET, no SKY either, set it to null string
             maskitdet = "" ! need both, so if no DET, no SKY either, set it to null string
          end if
       else
          write(errstr,*) &
               "SKY region file selected does NOT exist. Will NOT proceed with source exclusion."
          call message(AppMsg,Always)
          maskitsky = "" ! need both, so if no DET, no SKY either, set it to null string
          maskitdet = "" ! need both, so if no DET, no SKY either, set it to null string
       end if
    endif
    ! Step 6.
    ! open and read region file, update withregion if can't read/find regionfile
    ! SS: Get the definition of the object region
    fovdef = ""
    if (withreg) then 
       step = step + 1
       write(cstep,'(I2)')step
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Open and read region file: ',&
            trim(adjustl(inREG_F))
       call message(AppMsg,Always)
       call read_region(inREG_F,fovdef,withreg)
    endif
    !
    ! Step 11. Create an attitude file
    !
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Create an attitude (atthk.fits) file'
    call message(AppMsg,Always)
    go = .TRUE. ! fix this, low priority !!!???
    atthk_F = "atthk.fits" ! hardcode this???!!!
    if (go) then
       command = "atthkgen atthkset=" // trim(adjustl(atthk_F)) // " timestep=1"
       call system(trim(command))
    else
       write(errstr,*) "atthk.fits file exists will not rerun atthkgen."
       call message(AppMsg,Always)
    end if
    ! Step 7. Extract the corner region from observation data into event file
    ! isn't this done by espfilt, too???!!!
    ! docorner does the corner selections
    ! OLDNEW: mos1S001-corn.fits mos1S001-corevc.fits
    !
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Extract eventfile from corner region'
    call message(AppMsg,Always)
    call docorner(inEVLI_F,inCORN_F,ccddef,corndef,pattern)
    ! Step 8: Extract the entire FOV region from observation data into an image
    ! OLDNEW: mos1S001-clean mos1S001-allevc.fits
    ! OLDNEW: mos1S001-bkg_region-sky.fits mos1S001-bkgreghsky.fits
    ! OLDNEW: mos1S001-obj-im.fits mos1S001-fovimt.fits
    ! NEW mos1S001-allevc.fits + mos1S001-bkgreghsky.fits = mos1S001-fovimt.fits
    ! isn't this done by cheese???!!! also; how 'well' do the soft and hard
    ! 'bands' in cheese line up with what one would put in here as elow/ehigh???!!!
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Extract entire FOV region into an image'
    call message(AppMsg,Always)
    call makeImFile(inEVLI_F,inIMG_F,maskitsky,instE,ccddefa,pattern) 
    ! Step 9. Extract the entire FOV region from observation data into an image
    ! Do it again but use all events.  This will be a mask for calculating
    ! the SP scaling
    ! OLDNEW: mos1S001-obj-im-sp-det.fits/mos1S001-fovimspdet.fits
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Make mask for calculating the SP scaling'
    call message(AppMsg,Always)
    call makeSPfile(inEVLI_F,inSPMASK_F,maskitdet,fulldef,ccddef,fovdef,&
         celow,cehigh,pattern)
    ! Step 10. Extract the entire FOV region and band from observation data into an 
    ! image in detector coords (DET)
    ! OLDNEW: mos1S001-obj-im-det-350-1100.fits/mos1S001-fovimdet-350-1100.fits
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
         ': Extract entire FOV into (DET) img in selected E band'
    call message(AppMsg,Always) 
    inBAND_F = trim(adjustl(prefix)) // "-fovimdet-" // trim(adjustl(celow)) // &
         "-" // trim(adjustl(cehigh)) // ".fits" ! updated for consistency???!!!
    call makeBandImageDet (inEVLI_F,inBAND_F,maskitdet,fulldef,ccddef,ccddefa,fovdef,&
         celow,cehigh,pattern)
    !
    ! Step 12. Now make an exposure map for the FOV image
    ! inEXPMAP_F can be user selected!!!
    ! OLDNEW: mos1S001-clean.fits/mos1S001-allevc.fits, eventset
    ! OLDNEW: mos1S001-obj-im.fits/mos1S001-fovimt.fits imageset
    ! OLDNEW: mos1S001-exp-im.fits/mos1S001-expimt.fits expimageset
    go = .TRUE.
    if (go) then
       step = step + 1
       write(cstep,'(I2)')step
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            ': Make an exposure map for the FOV image (eexpmap)'
       call message(AppMsg,Always)
       command = "eexpmap attitudeset=atthk.fits eventset=" // &
            trim(adjustl(inEVLI_F)) // ":EVENTS " // &
            "expimageset=" // trim(adjustl(inEXPMAP_F)) // &
            " imageset=" // trim(adjustl(inIMG_F)) // &
            " pimin=300 pimax=8000 withdetcoords=no"
       call system(trim(command))
    else
       write(errstr,*) "FOV exposure map already exists will not re-run eexpmap here."
       call message(AppMsg,Always)
    end if
    ! Step 12 Make a mask
    ! OLDNEW: mos1S001-obj-im.fits mos1S001-fovim(h/s/t?).fits
    ! OLDNEW: mos1S001-mask-im.fits/mos1S001-maskim(h/s/t?).fits
    step = step + 1
    write(cstep,'(I2)')step
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Create a mask for the FOV image (emask)'
    call message(AppMsg,Always)
    write(errstr,*)trim(adjustl(inMASK_F))," ",trim(adjustl(inEXPMAP_F))
    call message(AppMsg,Always)
    go = .TRUE.
    if (go) then
       call run_emask(inMASK_F,inEXPMAP_F) !!! fix this !!!???
    else
       write(errstr,*) "FOV mask from FOV image exists, will not rerun emask here."
       call message(AppMsg,Always)
    end if
    !
    !  Step 13 Extract the entire FOV region from observation data into spectral file
    !  then run backscale, generate response
    !
    go = .TRUE.
    if (go) then ! fix this, low priority???!!!
       ! OLDNEW: mos1S001-clean.fits  mos1S001-allevc.fits
       ! OLDNEW: mos1S001-bkg_region-det.fits mos1S001-bkgreg(h/s/t)sky.fits
       ! OLDNEW: mos1S001-obj.pi  mos1S001-fov(h/s/t).pi
       step = step + 1
       write(cstep,'(I2)')step ! 13a
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'a: Extract FOV spectrum: ',&
            trim(adjustl(inSPEC_F))
       call message(AppMsg,Always) 
       command = "evselect table=" // trim(adjustl(inEVLI_F)) // &
            ":EVENTS withfilteredset=yes " // & 
            "expression='(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
            trim(adjustl(fovdef)) // trim(adjustl(ccddef)) // &
            trim(adjustl(maskitdet)) // trim(adjustl(fulldef)) // &
            "' filtertype=expression keepfilteroutput=no updateexposure=yes "// &
            "filterexposure=yes withspectrumset=yes spectrumset=" // &
            trim(adjustl(inSPEC_F)) // " energycolumn=PI " // & 
            "spectralbinsize=5 withspecranges=yes specchannelmin=0 " // &
            "specchannelmax=11999"
       call system(trim(command))
    else
       write(errstr,*) "FOV spec exists, will not rerun evselect here."
       call message(AppMsg,Always)
    end if
    ! Step 13.a run backscale
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // 'b: Backscale FOV spectrum'
    call message(AppMsg,Always) 
    command = "backscale spectrumset=" // trim(adjustl(inSPEC_F)) // &
         " badpixlocation=" // trim(adjustl(inEVLI_F)) // " withbadpixcorr=yes" // &
         " badpixelresolution=" // trim(adjustl(cpixRes))
    call system(trim(command))
    !
    ! Step 14. Generate the RMF
    !
    step = step + 1
    write(cstep,'(I2)')step ! 14
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Generate the RMF: ',&
         trim(adjustl(inRMF_F))
    call message(AppMsg,Always) 
    go = .TRUE.
    if (go) then
       command = "evselect table=" // trim(adjustl(inEVLI_F)) // &
            ":EVENTS withfilteredset=yes " // & 
            "expression='(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
            trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // &
            trim(adjustl(ccddef)) // trim(adjustl(maskitdet)) // &
            "' imagebinning='imageSize' imagedatatype='Int32' " // &
            "imageset=detmap.ds squarepixels=yes ignorelegallimits=no " // &
            "withxranges=yes withyranges=yes xcolumn='DETX' ximagesize=120 " // &
            "ximagemax=19500 ximagemin=-19499 ycolumn='DETY' yimagesize=120 " // &
            "yimagemax=19500 yimagemin=-19499 updateexposure=yes " // &
            "filterexposure=yes"
       call system(trim(command))
       command = "rmfgen format=var rmfset=" //trim(adjustl(inRMF_F)) // &
            " spectrumset=" // trim(adjustl(inSPEC_F)) // &
            " threshold=1.0e-6 detmaptype=dataset detmaparray=detmap.ds"
       call system(trim(command))
    end if
    !
    ! Step 15 Generate the ARF, use a detector map
    !
    step = step + 1
    write(cstep,'(I2)')step ! 15
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) // ': Generate the ARF: ',&
         trim(adjustl(inARF_F))
    call message(AppMsg,Always) 
    go = .TRUE.
    if (go) then
       call run_arfgen(inEVLI_F,maskitdet,inSPEC_F,inRMF_F,inARF_F,&
            fovdef,fulldef,ccddef,pattern)
    end if
    ! Step 16 make SKY coord diagnostic images for the bands
    ! OLDNEW: mos1S001-clean.fits mos1S001-fovevc.fits
    ! OLDNEW: mos1S001-bkg_region-det.fits mos1S001-bkgregtsky.fits
    ! OLDNEW: mos1S001-obj-im-350-1100.fits mos1S001-fovimsky-350-1100.fits
       inBAND_F = trim(adjustl(prefix)) // "-fovimsky-" // trim(adjustl(celow)) // &
            "-" // trim(adjustl(cehigh)) // ".fits"
       step = step + 1
       write(cstep,'(I2)')step ! 16
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "a: Creating FOV selected E band image (SKY): " // trim(adjustl(inBAND_F))
       call message(AppMsg,Always) 
       ! Step 16.a
       ! OLDNEW: mos1S001-obj-im-elow-ehigh.fits/mos1S001-fovimsky-elow-ehigh.fits
       call makeBandImageSky (inEVLI_F,inBAND_F,maskitsky,fulldef,ccddef,ccddefa,fovdef,&
            celow,cehigh,pattern)
       ! Make an exposure map for the region and band
       ! OLDNEW: mos1S001-expim-350-1100.fits/mos1S001-expimsky-350-1100.fits
       inBAND_F = trim(adjustl(prefix)) // "-expimsky-" // trim(adjustl(celow)) // &
            "-" // trim(adjustl(cehigh)) // ".fits"
       go = .TRUE. ! fix this???!!!
       ! Step 16.b
       if (go) then
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "b: Make an exposure map for the region and selected E band (eexpmap)"
          call message(AppMsg,Always) 
          command = "eexpmap attitudeset=atthk.fits eventset=" // &
               trim(adjustl(inEVLI_F)) // ":EVENTS " // &
               "expimageset=" // trim(adjustl(inBAND_F)) // &
               " imageset=" // trim(adjustl(inIMG_F)) // &
               " pimax=" // trim(adjustl(cehigh)) // &
               " pimin=" // trim(adjustl(celow)) // " withdetcoords=no"
          call system(trim(command))
       end if
       ! Make a mask for the region and the band 16.c
       ! OLDNEW: mos1S001-mask-im-350-1100.fits mos1S001-maskim-350-1100.fits
       ! OLDNEW: uses mos1S001-fovim(h/s/t).fits 
       inMASK_F = trim(adjustl(prefix)) // "-maskimsky-" // trim(adjustl(celow)) // &
            "-" // trim(adjustl(cehigh)) // ".fits"
       !       call gonogo(inMASK_F,go)
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "c: Make a mask for the region and selected E band (emask)"
       call message(AppMsg,Always) 
       go = .TRUE.
       if (go) then
          call run_emask(inMASK_F,inBAND_F)
       end if
    ! open inSPEC_F (e.g. mos1S001-fovt.pi) to get keywords to add to FWC spectra below
    ! stays open until all FWC specs are filled (08/25/2021)
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inSPEC_F),inSPEC_S,1,status,warndie)
    !
    !  SS: Extract region spectra separately for each chip
    !
    !  SS: For chip 1 just get object region from FWC and backscale that 
    ! 
    if (ccds(1)) then
       step = step + 1
       write(cstep,'(I2)')step ! 17a
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "a:  Extract spectrum from FWC CCF for ccd1"
       call message(AppMsg,Always)
       ! Step 17.a
       ! OLDNEW: mos1S001-1ff.pi mos1S001-fovfwccd1.pi 
       ccdSPEC_F = trim(adjustl(prefix)) // "-fovfwcccd1.pi"
       command = "evselect table=" // trim(adjustl(inFWC_F)) // ":EVENTS " // &
            "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
            "&&(FLAG == 0)&&(CCDNOISE<=1)&&" // &
            trim(adjustl(chipdef(1))) // trim(adjustl(ccddef)) // &
            trim(adjustl(maskitdet)) // trim(adjustl(fulldef)) // &
            trim(adjustl(fovdef)) // "' withspectrumset=yes " // &
            "spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
            " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
            "specchannelmin=0 specchannelmax=11999"
       call system(trim(command))
       ! put header keywords from *fovt.pi into *fovfwcccd1.pi
       ! corrects backscale failure for not finding keys in FWC CCF (BHP, 08/25/2021)
       status = 0
       warndie = .TRUE.
       call check_open_set(trim(ccdSPEC_F),ccdSPEC_S,4,status,warndie) ! 4='open to write'
       write(errstr,*)  " **** Copying these attributes: INSTRUME TELESCOP DATE-OBS DATE-END " &
            // "FILTER RA_PNT DEC_PNT EXPOSURE"
       call message(AppMsg,Verbose)
       write(errstr,*)  "from ",trim(adjustl(inSPEC_F)),' to ',trim(adjustl(ccdSPEC_F))
       call message(AppMsg,Verbose)
       call copyARFattributes(inSPEC_S,ccdSPEC_S,expo_CCD(1))
       call release(ccdSPEC_S)
       ! BHP: new for SAS-20, up cpixRes to speed processing (default=16) 17b
       command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
            " badpixlocation=" // trim(adjustl(inEVLI_F)) // " withbadpixcorr=yes " // &
            " badpixelresolution=" // trim(adjustl(cpixRes))
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "b:  Backscale spectra from FWC CCF for ccd1"
       call message(AppMsg,Always)
       call system(trim(command))
       ! orig: 17c
       ! SS: Make an image of the FWC data in a band for the central chip
       ! BHP: if withimages = .FALSE. mosspectra only makes spectra, no images
          ! 17.c
          ! SS: Need to take care of the window mode case for the central CCD
          !  First make an image of the observation data for the central chip
          !
          ! OLDNEW: mos1S002-obj-im-350-1100-ccd1.fits mos1S001-fovimccd1-350-1100.fits
          !
          ccdIMG_F = trim(adjustl(prefix)) // "-fovimccd1-" // trim(adjustl(celow)) // &
               "-" // trim(adjustl(cehigh)) // ".fits"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "c: Extract image from FOV (window mode case) in selected E band for ccd1 "
          call message(AppMsg,Always)
          command = "evselect table=" // trim(adjustl(inEVLI_F)) // &
               ":EVENTS withfilteredset=yes " // & 
               "expression='"// trim(adjustl(chipdef(1))) // &
               "&&(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
               trim(adjustl(ccddefa)) // &
               trim(adjustl(maskitdet)) // "&&(PI in [" // trim(adjustl(celow)) // &
               ":" // trim(adjustl(cehigh)) // "])" // trim(adjustl(fulldef)) // &
               trim(adjustl(fovdef)) //  "' filtertype=expression imagebinning='imageSize'" // &
               " imagedatatype='Int32' imageset=" // trim(adjustl(ccdIMG_F)) // &
               " squarepixels=yes withxranges=yes withyranges=yes xcolumn='DETX' " // &
               "ximagesize=780 ximagemax=19500 ximagemin=-19499 ycolumn='DETY' " // &
               "yimagesize=780 yimagemax=19500 yimagemin=-19499 updateexposure=yes " // &
               "filterexposure=yes ignorelegallimits=yes "
          call system(trim(command))
          !
          ! add exposure calculated above from FWC EXPTIME extension
          ! do this???!!!
          ! call check_open_set(trim(ccdIMG_F),ccdIMG_S,4,status,warndie)
          ! call setAttribute(ccdIMG_S,'EXPOSURE',expo_CCD(1),'Exposure Time','s')
          ! call release(ccdIMG_S)
          !
          ! SS: Next make an exposure map for the region and band for the central CCD
          ! Step 17.d
          ! OLDNEW: mos1S002-exp-im-350-1100-ccd1.fits mos1S002-expimccd1-350-1100.fits
          ccdEXPIMG_F = trim(adjustl(prefix)) // "-expimccd1-" // trim(adjustl(celow)) // &
               "-" // trim(adjustl(cehigh)) // ".fits"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "d: Make exposure map from FOV in selected E band (eexpmap) for ccd1"
          call message(AppMsg,Always)
          command = "eexpmap attitudeset=atthk.fits eventset=" // &
               trim(adjustl(inEVLI_F)) // ":EVENTS " // &
               "expimageset=" // trim(adjustl(ccdEXPIMG_F)) // &
               " imageset=" // trim(adjustl(ccdIMG_F)) // &
               " pimax="  // trim(adjustl(cehigh)) // &
               " pimin=" // trim(adjustl(celow)) // " withdetcoords=yes"
          call system(trim(command))
          !
          ! Next make a mask for the region and the band for the central CCD
          ! OLDNEW: mos1S001-mask-im-350-1100-ccd1.fits mos1S001-maskimccd1-350-1100.fits
          ! orig: 17.e
          ccdMASK_F = trim(adjustl(prefix)) // "-maskimccd1-" // &
               trim(adjustl(celow)) // "-" // trim(adjustl(cehigh)) // ".fits"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "e: Make mask from FOV in selected E band (emask) for ccd1"
          call message(AppMsg,Always)
          call run_emask(ccdMASK_F,ccdEXPIMG_F) !!! check thresholds!!! gonogo???
          ! orig: 17.f
          ! SS: Make an image of the fwc data in a band for the central chip
          ! OLDNEW: mos1S001-im1-350-1100.fits mos1S001-fwcimccd1-350-1100.fits
          fwcIMG_F = trim(adjustl(prefix)) // "-fwcimccd1det-" // &
               trim(adjustl(celow)) // &
               "-" // trim(adjustl(cehigh)) // ".fits"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "f: Extract FWC image in selected E band for ccd1"
          call message(AppMsg,Always)
          command = "evselect table=" // trim(adjustl(inFWC_F)) // &
               ":EVENTS withfilteredset=yes " // & 
               "expression='(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
               "&&(CCDNOISE<=1)" // trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // &
               trim(adjustl(ccddef)) // trim(adjustl(maskitdet)) // "&&" // &
               trim(adjustl(chipdef(1))) // "&&(PI in [" // trim(adjustl(celow)) // &
               ":" // trim(adjustl(cehigh)) // "])' imagebinning='imageSize' " // &
               "imagedatatype='Int32' " // "imageset=" // trim(adjustl(fwcIMG_F)) // &
               " squarepixels=yes withxranges=yes withyranges=yes xcolumn='DETX' " // &
               "ximagesize=780 ximagemax=19500 ximagemin=-19499 ycolumn='DETY' " // &
               "yimagesize=780 yimagemax=19500 yimagemin=-19499 updateexposure=yes " // &
               "filterexposure=yes ignorelegallimits=yes "
          call system(trim(command))
          ! call check_open_set(trim(fwcIMG_F),fwcIMG_S,4,status,warndie)
          ! call setAttribute(fwcIMG_S,'EXPOSURE',expo_CCD(1),'Exposure Time','s')
          ! call release(fwcIMG_S) do this???!!!
          ! 
          ! Mask the FWC image (fix this with SAS code!!!???)
          ! orig: 17.g fix this to NOT use farith in SAS-21???!!!
          tempout_F = trim(adjustl(prefix)) // "-fwcmaskccd1-"// &
               trim(adjustl(celow)) // "-" // trim(adjustl(cehigh)) // ".fits"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "g: Mask the FWC image (farith) ccd1"
          call message(AppMsg,Always)
          command = "farith infil1='" // trim(adjustl(fwcIMG_F)) // "' " // &
               "infil2='" // trim(adjustl(ccdMASK_F)) // "[MASK]' " // &
               "outfil='" // trim(adjustl(tempout_F)) // "' ops=MUL " // &
               " clobber=yes copyprime=yes"
          call system(trim(command))
          ! Rename the FWC image why???!!!
          status = 0
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "h: Rename FWC mask ccd1"
          call message(AppMsg,Always)
          call rename(tempout_F,fwcIMG_F,status) ! check this???!!!
          if (status.ne.0) then
             call error("noRENAME","Failed to rename FWC file to IMG filename!")
          end if
       ! orig: 17.h
       ! SS: Make a spectrum of the observation data for the central chip
       ! OLDNEW: mos1S001-1obj.pi mos1S001-fovccd1.pi
       !
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "i:  Extract FOV spectrum for ccd1"
       call message(AppMsg,Always)
       ccdSPEC_F = trim(adjustl(prefix)) // "-fovccd1.pi"
       command = "evselect table=" // trim(adjustl(inEVLI_F)) // &
            ":EVENTS withfilteredset=yes " // &
            "expression='" // trim(adjustl(chipdef(1))) // &
            "&&(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
            trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // &
            trim(adjustl(ccddef)) // trim(adjustl(maskitdet)) // &
            "' filtertype=expression keepfilteroutput=no updateexposure=yes " // &
            "spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
            " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
            "specchannelmin=0 specchannelmax=11999"
       write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
            "j:  Backscale FOV spectrum for ccd1"
       call message(AppMsg,Always)
       call system(trim(command))
       command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
            " badpixlocation="// trim(adjustl(inEVLI_F)) // " withbadpixcorr=yes" // &
            " badpixelresolution=" // trim(adjustl(cpixRes))
       call system(trim(command))
    endif
    ! orig: 18
    ! SS: Now do the remainder of chips
    !
    step = step + 1
    write(cstep,'(I2)')step ! 17
    do i=2,7
       if (ccds(i)) then
          write(schip,'(I1)')i
          ! Extract a spectrum from the corner pixel data from the observation, 
          ! then backscale it
          ! Step 20.a
          ! OLDNEW: mos1S001-corn.fits mos1S001-corevc.fits
          ! OLDNEW: mos1S001-2oc.pi mos1S001-corccd2.pi
          ccdSPEC_F = trim(adjustl(prefix)) // "-corccd" // schip // ".pi"
          go = .TRUE.
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "a:  Extract CORNER spectra for ccd" // schip
          call message(AppMsg,Always)
          if (go) then
             command = "evselect table=" // trim(adjustl(inCORN_F)) // ":EVENTS " // &
                  "withfilteredset=yes expression='" // trim(adjustl(chipdef(i))) // &
                  "' filtertype=expression keepfilteroutput=no " // &
                  "withspectrumset=yes spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
                  " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
                  "specchannelmin=0 specchannelmax=11999"
             call system(trim(command))
          end if
          ! orig: 18.b
          command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
               " badpixlocation="// trim(adjustl(inEVLI_F)) // &
               " withbadpixcorr=yes ignoreoutoffov=no"  // &
               " badpixelresolution=" // trim(adjustl(cpixRes))
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "b: Backscale CORNER spectra for ccd" // schip
          call message(AppMsg,Always)
          call system(trim(command))
          ! orig: 18.c
          ! Step 20.c Extract a spectrum of the observation data from the FOV
          ! OLDNEW: mos1S001-2obj.pi mos1S001-fovccd2.pi
          ccdSPEC_F = trim(adjustl(prefix)) // "-fovccd"  // schip // ".pi"
          go = .TRUE.
          if (go) then
             write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
                  "c: Extract FOV spectra for ccd" // schip
             call message(AppMsg,Always)
             command = "evselect table=" // trim(adjustl(inEVLI_F)) // ":EVENTS " // &
                  "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
                  "&&(FLAG == 0)" // &
                  trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // &
                  "&&" // trim(adjustl(chipdef(i))) // trim(adjustl(ccddef)) // &
                  trim(adjustl(maskitdet)) // "' filtertype=expression " // &
                  "keepfilteroutput=no updateexposure=yes filterexposure=yes " // &
                  "withspectrumset=yes spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
                  " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
                  "specchannelmin=0 specchannelmax=11999"
             call system(trim(command))
          endif
          ! run backscale even if exists??? orig: 18.d
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "d: Backscale FOV spectra for ccd" // schip
          call message(AppMsg,Always)
          command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
               " badpixlocation="// trim(adjustl(inEVLI_F)) // &
               " withbadpixcorr=yes " // &
               " badpixelresolution=" // trim(adjustl(cpixRes))
          call system(trim(command))
          !
          ! Step 20.e Set up for the FWC data with selection of appropriate files
          ! Determine the rate
          ! orig: 18.e
          inquire(FILE="temp_events.fits", EXIST=i_ok)
          if (i_ok) then
             call unlink ("temp_events.fits")
          endif
          command = "evselect table=" // trim(adjustl(inCORN_F)) // ":EVENTS " // &
               "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
               "&&((FLAG & 0x766a0f63)==0)&&" // &
               trim(adjustl(chipdef(i))) // "&&(PI in [300:10000])' " // &
               " filtertype=expression " // &
               "filteredset=temp_events.fits keepfilteroutput=yes " // &
               "updateexposure=yes filterexposure=yes "
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "e:  Create temp event list from FOV data ccd"//schip
          call message(AppMsg,Always)
          call system(trim(command))
          !
          ! OLD fkeypar section (call read_header_info in gofio instead)
          ! get these from new temp_events.fits NAXIS1, expo, LIVETInn
          ! Step 20.f
          status = 0
          warndie = .TRUE.
          call check_open_set("temp_events.fits",temp_S,1,status,warndie)
          call getkeys(temp_S,cnts,expo,livetime,schip)
          call release(temp_S)
          call unlink ("temp_events.fits")
          ! Step 20.g
          rate = (cnts/livetime) * 100.
          !
          ! Determine the hardness ratio
          ! orig: 18.h
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "f:  Determine the hardness ratio ccd"//schip
          call message(AppMsg,Always)
          command = "evselect table=" // trim(adjustl(inCORN_F)) // ":EVENTS " // &
               "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
               "&&((FLAG & 0x766a0f63)==0)&&" // trim(adjustl(chipdef(i))) // &
               "&&(PI in [500:800])' filtertype=expression " // &
               "filteredset=temp_events.fits keepfilteroutput=yes " // &
               "updateexposure=yes filterexposure=yes "
          call system(trim(command))
          ! return lowe (NAXIS2), expo, LIVETInn
          status = 0
          warndie = .TRUE.
          call check_open_set("temp_events.fits",temp_S,1,status,warndie)
          ! Step 20.i
          call getkeys(temp_S,ilo,expo,livetime,schip)
          call release(temp_S)
          call unlink ("temp_events.fits")   
          command = "evselect table=" // trim(adjustl(inCORN_F)) // ":EVENTS " // &
               "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
               "&&((FLAG & 0x766a0f63)==0)&&" // trim(adjustl(chipdef(i))) // &
               "&&(PI in [2500:5000])' filtertype=expression " // &
               "filteredset=temp_events.fits keepfilteroutput=yes " // &
               "updateexposure=yes filterexposure=yes "
          call system(trim(command))
          ! return hige (NAXIS2), expo, LIVETInn
          call check_open_set("temp_events.fits",temp_S,1,status,warndie)
          ! Step 20.k
          call getkeys(temp_S,ihi,expo,livetime,schip)
          call release(temp_S)
          call unlink ("temp_events.fits")   
          write(errstr,*) " Hardness Ratio: "
          call message(AppMsg,Always)
          ! Step 20.l
          hige = real(ihi)
          lowe = real(ilo)
          if ((lowe .gt. 0) .and. (hige .gt. 0)) then 
             hard = hige/lowe
          else
             write(errstr,*)"Zero events for CCD",schip," while calculating Hardness ratio"
             call message(AppMsg,Always)
             write(errstr,*)"Try re-running mosspectra with CCD",schip," deselected"
             call error("ZeroEvents",trim(adjustl(errstr)))
          endif
          ! Step 20.m
          ! orig: 18.l
          hardsig=sqrt((hige/(lowe*lowe))+((hige*hige*lowe)/(lowe*lowe*lowe*lowe)))
          write(errstr,*)  " hige, lowe, hard, hardsig = ",hige,lowe,hard,hardsig
          call message(AppMsg,Always)
          suff = ""
          write(errstr,*)  " Corner rate     = ",rate
          call message(AppMsg,Always)
          write(errstr,*)  " Corner hardness = ",hard
          call message(AppMsg,Always)
          !
          ! anom code stays until checked vs SOC fwc data column, then use that, or no???!!!
          !          
          ! SS: Determine which FWC data file is required
          ! SS: Currently this defaults to the "good" parts version
          ! SS: until considerably more testing can be done
          ! orig: 18.n.
          ! BHP: Long calculations moved to subroutine anom. Put this in emanom code as a
          ! callable subroutine for sas-21???!!!
          call anom(instE,i,rate,hard,hardsig,ccds,newccds)
          ! orig: 18.o
          !  SS: Extract a spectrum from the corner pixel data from the FWC data, 
          !  SS: then backscale it
          ! OLDNEW: mos1S001-2fc.pi mos1S001-corfwcccd2.pi
          ccdSPEC_F = trim(adjustl(prefix)) // "-corfwcccd"  // schip // ".pi"
          go = .TRUE. ! fix low priority!!!???
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "g:  Extract FWC CORNER spectrum for ccd"//schip
          call message(AppMsg,Always)
          if (go) then
             command = "evselect table=" // trim(adjustl(inFWC_F)) // ":EVENTS " // &
                  "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
                  "&&(FLAG == 65536)" // "&&(CCDNOISE<=1)&&" // trim(adjustl(chipdef(i))) // &
                  trim(adjustl(corndef)) // "' filtertype=expression keepfilteroutput=no " // &
                  "keepfilteroutput=no updateexposure=yes filterexposure=yes " // &
                  "withspectrumset=yes spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
                  " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
                  "specchannelmin=0 specchannelmax=11999"
             call system(trim(command))
          end if
          ! put header keywords from *fovt.pi into *corfwcccd1.pi
          ! corrects backscale failure for not finding keys in FWC CCF (BHP, 08/25/2021)
          status = 0
          warndie = .TRUE.
          call check_open_set(trim(ccdSPEC_F),ccdSPEC_S,4,status,warndie) ! 4='open to write'
          write(errstr,*)  " **** Copying these attributes: INSTRUME TELESCOP DATE-OBS DATE-END " &
               // "FILTER RA_PNT DEC_PNT EXPOSURE"
          call message(AppMsg,Verbose)
          write(errstr,*)  "from ",trim(adjustl(inSPEC_F)),' to ',trim(adjustl(ccdSPEC_F))
          call message(AppMsg,Verbose)          
          call copyARFattributes(inSPEC_S,ccdSPEC_S,expo_CCD(i))
          call release(ccdSPEC_S)
          !
          ! run backscale even if exists???
          ! orig: 18.q
          command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
               " badpixlocation="// trim(adjustl(inEVLI_F)) // &
               " withbadpixcorr=yes ignoreoutoffov=no "  // &
               " badpixelresolution=" // trim(adjustl(cpixRes))
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "h:  Backscale FWC CORNER spectrum from ccd"//schip
          call message(AppMsg,Always)
          call system(trim(command))
          !
          ! SS: Extract a spectrum from the FOV region from the FWC data, backscale it
          ! OLDNEW: mos1S001-2ff.pi mos1S001-fovfwcccd2.pi
          ! orig: 18.r
          ccdSPEC_F = trim(adjustl(prefix)) // "-fovfwcccd" // schip // ".pi"
          command = "evselect table=" // trim(adjustl(inFWC_F)) // ":EVENTS " // &
               "withfilteredset=yes expression='(FLAG == 0)" // &
               "&&(CCDNOISE<=1)" // trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // "&&" // &
               trim(adjustl(chipdef(i))) // trim(adjustl(ccddef)) // &
               trim(adjustl(maskitdet)) // "' withspectrumset=yes " // &
               "keepfilteroutput=no updateexposure=yes filterexposure=yes " // &
               "spectrumset=" // trim(adjustl(ccdSPEC_F)) // &
               " energycolumn=PI spectralbinsize=5 withspecranges=yes " // &
               "specchannelmin=0 specchannelmax=11999"
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "i:  Extract FWC FOV spectrum for ccd"//schip
          call message(AppMsg,Always)
          call system(trim(command))
          ! put header keywords from *fovt.pi into *fovfwcccd1.pi
          ! corrects backscale failure for not finding keys in FWC CCF (BHP, 08/25/2021)
          status = 0
          warndie = .TRUE.
          call check_open_set(trim(ccdSPEC_F),ccdSPEC_S,4,status,warndie) ! 4='open to write'
          write(errstr,*)  " **** Copying these attributes: INSTRUME TELESCOP DATE-OBS DATE-END " &
               // "FILTER RA_PNT DEC_PNT EXPOSURE"
          call message(AppMsg,Verbose)
          write(errstr,*)  "from ",trim(adjustl(inSPEC_F)),' to ',trim(adjustl(ccdSPEC_F))
          call message(AppMsg,Verbose)                    
          call copyARFattributes(inSPEC_S,ccdSPEC_S,expo_CCD(i))
          call release(ccdSPEC_S)
          !
          command = "backscale spectrumset=" //  trim(adjustl(ccdSPEC_F)) // &
               " badpixlocation="// trim(adjustl(inEVLI_F)) // &
               " withbadpixcorr=yes"  // &
               " badpixelresolution=" // trim(adjustl(cpixRes))
          write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
               "j:  Backscale FWC FOV spectrum for ccd"//schip
          call message(AppMsg,Always)
          call system(trim(command))
          !
          ! SS: Make an image of the fwc data in the input band
          ! orig: 18.s
          ! BHP: if withimages = .FALSE. mosspectra only makes spectra, no images
             ccdIMG_F = trim(adjustl(prefix)) // "-fwcimccd"  // schip // &
                  "det" // "-" // trim(adjustl(celow)) // "-" // &
                  trim(adjustl(cehigh)) // ".fits"
             command = "evselect table=" // trim(adjustl(inFWC_F)) // ":EVENTS " // &
                  "withfilteredset=yes expression='" // trim(adjustl(chipdef(i))) // &
                  trim(adjustl(ccddef)) // trim(adjustl(maskitdet)) // &
                  "&&(CCDNOISE<=1)&&(FLAG == 0)&&(PI in [" // trim(adjustl(celow)) // ":" // &
                  trim(adjustl(cehigh)) // "])" // trim(adjustl(fulldef)) // &
                  trim(adjustl(fovdef)) // "' imagebinning='imageSize' " // &
                  "imagedatatype='Int32' " // &
                  "imageset=" // trim(adjustl(ccdIMG_F)) // " squarepixels=yes " // &
                  "withxranges=yes withyranges=yes xcolumn='DETX' ximagesize=780 " // &
                  "ximagemax=19500 ximagemin=-19499 ycolumn='DETY' yimagesize=780 " // &
                  "yimagemax=19500 yimagemin=-19499 updateexposure=yes filterexposure=yes " // &
                  "ignorelegallimits=yes "
             write(errstr,*) 'Step ' // trim(adjustl(cstep)) // &
                  "k:  Extract FWC FOV image in the selected E band for ccd" // schip
             call message(AppMsg,Always)
             call system(trim(command))
             !
             ! add exposure calculated above from FWC EXPTIME extension
             ! do this???!!!
!             call check_open_set(trim(ccdIMG_F),ccdIMG_S,4,status,warndie)
!             call setAttribute(ccdIMG_S,'EXPOSURE',expo_CCD(1),'Exposure Time','s')
!             call release(ccdIMG_S)
             ! why no mask, farith, rename like in pnspectra???!!!
       end if
    end do
    call release(inSPEC_S) ! input file used to transfer keywords to FWC products
    !
    ! clean up (add things here!!!)
    ! Step 21
    step = step + 1
    write(cstep,'(I2)')step ! 21
    write(errstr,*) 'Step ' // trim(adjustl(cstep)) //&
         ": mosspectra analysis complete; cleaning up..."
    call message(AppMsg,Always)
    if (keep) then
       write(errstr,*) "MOSSPECTRA: parameter keepinterfiles IS set."
       call message(AppMsg,Always)
       write(errstr,*) "MOSSPECTRA: Will NOT delete these files (also produced by other tasks):"
       call message(AppMsg,Always)
       write(errstr,*) "MOSSPECTRA: detmap.ds, filtered.fits, temp_events.fits"
       call message(AppMsg,Always)
    else       
       write(errstr,*)"Cleaning interfiles..."
       call message(AppMsg,Always)
       write(errstr,*) "MOSSPECTRA: keepinterfiles NOT set"
       call message(AppMsg,Always)
       write(errstr,*) "MOSSPECTRA: Deleting intermediary files..."
       call message(AppMsg,Always)
       write(errstr,*) "Deleting detmap.ds"
       call message(AppMsg,Always)
       call unlink("detmap.ds")
       write(errstr,*) "Deleting filtered.fits"
       call message(AppMsg,Always)
       call unlink("filtered.fits")
       write(errstr,*) "Deleting temp_events.fits"
       call message(AppMsg,Always)
       call unlink("temp_events.fits")
    end if

    return

  end subroutine mosspectra
  !
  !
  !
  subroutine run_emask(detmask,expimg)
    !
    ! Makes masks files
    !
    character(len=*), intent(in) :: detmask, expimg
    character(len=1024) :: command
    ! add check_clobber???!!!
    command = "emask detmaskset=" // trim(adjustl(detmask)) // & 
         " expimageset=" // trim(adjustl(expimg)) // & 
         " threshold1=0.01 threshold2=0.5"
    call system(trim(command))
    return
    
  end subroutine run_emask
  !
  !
  !
  subroutine run_arfgen(evli,maskit,spec,rmf,arf,fovdef,fulldef,ccddef,pattern)
    !
    character(len=*) :: evli,maskit,spec,rmf,arf,fovdef,fulldef,ccddef,pattern
    character(len=4096) :: command
    ! add check_clobber???!!!

    write(errstr,*) "RUN_ARFGEN: Running evselect, then arfgen..."
    call message(AppMsg,Always)
    command = "evselect table=" // trim(adjustl(evli)) // &
         ":EVENTS withfilteredset=yes " // & 
         "expression='(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
         trim(adjustl(fovdef)) // trim(adjustl(fulldef)) // &
         trim(adjustl(ccddef)) // trim(adjustl(maskit)) // &
         "' imagebinning='imageSize' imagedatatype='Int32' " // &
         "imageset=detmap.ds squarepixels=yes ignorelegallimits=no " // &
         "withxranges=yes withyranges=yes xcolumn='DETX' ximagesize=120 " // &
         "ximagemax=19500 ximagemin=-19499 ycolumn='DETY' yimagesize=120 " // &
         "yimagemax=19500 yimagemin=-19499 updateexposure=yes filterexposure=yes"
   ! write(errstr,*) 'Step 15.a ',trim(command)
   ! call message(AppMsg,Always)
    call system(trim(command))
    command = "arfgen arfset=" // trim(adjustl(arf)) // " spectrumset=" // &
         trim(adjustl(spec)) //  " withrmfset=yes rmfset=" // trim(adjustl(rmf)) // &
         " extendedsource=yes modelee=no withbadpixcorr=no detmaptype=dataset " // &
         "detmaparray=detmap.ds badpixlocation=" // trim(adjustl(evli)) // &
         " modelootcorr=no"
   ! write(errstr,*) 'Step 15.b ',trim(command)
   ! call message(AppMsg,Always)
    call system(trim(command))

    return

  end subroutine run_arfgen
  !
  !
  !

  subroutine printParams(inEVLI_F,inCORN_F,inIMG_F,inSPMASK_F,inEXPMAP_F,inMASK_F,&
       withsrcX,inSKY_F,inDET_F,withreg,inREG_F,inRMF_F,inARF_F,ccds,&
       elow,ehigh,keep,pattern)

    character(len=*) :: inEVLI_F,inCORN_F,inIMG_F,inSPMASK_F,inEXPMAP_F,inMASK_F,&
         inSKY_F,inDET_F,inREG_F,inRMF_F,inARF_F
    integer :: elow, ehigh, i, pattern
    logical :: withsrcX, withreg, keep, ccds(7)

    write(errstr,*) "Input Cleaned FOV event list:    ",trim(adjustl(inEVLI_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input Cleaned CORNER event list: ",trim(adjustl(inCORN_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input Cleaned IMAGE:             ",trim(adjustl(inIMG_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input SP scaling mask:           ",trim(adjustl(inSPMASK_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input Exposure Map:              ",trim(adjustl(inEXPMAP_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input Band and Region mask :     ",trim(adjustl(inMASK_F))
    call message(AppMsg,Always)
    if (withsrcX) then
       write(errstr,*) "Input SKY coord region mask :    ",trim(adjustl(inSKY_F))
       call message(AppMsg,Always)
       write(errstr,*) "Input DET coord region mask :    ",trim(adjustl(inDET_F))
       call message(AppMsg,Always)
    endif
    if (withreg) then
       write(errstr,*) "Input text region file:          ",trim(adjustl(inREG_F))
       call message(AppMsg,Always)
    endif
    write(errstr,*) "Input Response Matrix (RMF) file:     ",trim(adjustl(inRMF_F))
    call message(AppMsg,Always)
    write(errstr,*) "Input Aux Response Matrix (ARF) file: ",trim(adjustl(inARF_F))
    call message(AppMsg,Always)
    write(errstr,*) "CCDs selected: ",ccds
    call message(AppMsg,Always)
    write(errstr,*) "Low energy threshold (eV):  ",elow
    call message(AppMsg,Always)
    write(errstr,*) "High energy threshold (eV): ",ehigh
    call message(AppMsg,Always)
    write(errstr,*) " evselect PATTERN (usually 12) : ",pattern
    call message(AppMsg,Always)
    write(errstr,*) "Keep intermediary files?: ",keep
    call message(AppMsg,Always)
    
    return

  end subroutine printparams
  !
  ! Define all the selection regions for evselect, etc
  !
  subroutine defineCCDregion (inst,ccds,ccddef,ccddefa)

    character(len=2), intent(in) :: inst
    character(len=1024), intent(inout) :: ccddef, ccddefa
    character(len=64) :: def(7), defa(7)
    integer :: cc, i, j
    logical, intent(in) :: ccds(7)
    
    def     = ""
    ccddef  = ""
    defa(1) =  "(CCDNR == 1)"
    defa(2) =  "(CCDNR == 2)"
    defa(3) =  "(CCDNR == 3)"
    defa(4) =  "(CCDNR == 4)"
    defa(5) =  "(CCDNR == 5)"
    defa(6) =  "(CCDNR == 6)"
    defa(7) =  "(CCDNR == 7)"
    if (count(ccds) .gt. 1) then
       ccddefa = "&&("
    end if
    !
    j = 0
    if (trim(adjustl(inst)) .eq. 'M1') then 
       write (errstr,*) "Defining special corner selection expression for EMOS1"
       def(1) = "&&!((DETX,DETY) in BOX(20,-40,6610,6590,0))"
       def(2) = "&&!((DETX,DETY) in BOX(6590,-13572,6590,6599,0))"
       def(3) = "&&!((DETX,DETY) in BOX(13280,-306,6610,6599,0))"
       def(4) = "&&!((DETX,DETY) in BOX(6630,13060,6500,6550,0))"
       def(5) = "&&!((DETX,DETY) in BOX(-6470,13130,6550,6560,0))"
       def(6) = "&&!((DETX,DETY) in BOX(-13169,-105,6599,6599,0))"
       def(7) = "&&!((DETX,DETY) in BOX(-6580,-13438,6570,6599,0))"
    elseif(inst .eq. 'M2') then
       write (errstr,*) "Defining special corner selection expression for EMOS2"
       def(1) = "&&!((DETX,DETY) in BOX(20,-40,6610,6590,0))"
       def(2) = "&&!((DETX,DETY) in BOX(6590,-13520,6630,6650,0))"
       def(3) = "&&!((DETX,DETY) in BOX(13320,-295,6620,6590,0))"
       def(4) = "&&!((DETX,DETY) in BOX(6610,13110,6590,6550,0))"
       def(5) = "&&!((DETX,DETY) in BOX(-6560,13180,6590,6600,0))"
       def(6) = "&&!((DETX,DETY) in BOX(-13190,-90,6600,6630,0))"
       def(7) = "&&!((DETX,DETY) in BOX(-6620,-13438,6620,6599,0))"
       !
    endif
    call message(AppMsg,Noisy)
    do i=1,7
       if (.not.ccds(i)) then
          ! cut out the non-selected ccds
          ccddef = trim(adjustl(ccddef)) // trim(adjustl(def(i)))
       else
          j = j + 1
          if (j .eq. 1) then
             ccddefa = "&&(" // trim(adjustl(defa(i)))
          else
             ccddefa = trim(adjustl(ccddefa)) // "||" // trim(adjustl(defa(i)))
          end if
       endif
    end do
    ! Close the region description
    ccddefa = trim(adjustl(ccddefa)) // ")"
!    write(errstr,*) &
!         "CCD region selection (ccddef) is:"
!    call message(AppMsg,Always)
!    write(errstr,*) trim(adjustl(ccddef))
!    call message(AppMsg,Always)
!    write(errstr,*) &
!         "CCD region selection (ccddefa) is:"
!    call message(AppMsg,Verbose)
!    write(errstr,*) trim(adjustl(ccddefa))
!    call message(AppMsg,Verbose)

    return

  end subroutine defineCCDregion
  !
  !
  !
  subroutine defineexpr (inst,cornerdef,chipdef,fulldef)

    character(len=5), intent(in) :: inst
    character(len=80) :: command
    character(len=*), intent(inout) :: cornerdef, fulldef, chipdef(7)
    cornerdef = ''
    chipdef   = ''
    fulldef   = ''
    !    
    if (inst .eq. 'EMOS1') then ! EMOS1
       write (errstr,*) "Defining corner selection expression for EMOS1"
       call message(AppMsg,Noisy)
       cornerdef = "&&!(((DETX,DETY) in CIRCLE(100,-200,17700))" // &
            "||((DETX,DETY) in CIRCLE(834,135,17100))" // &
            "||((DETX,DETY) in CIRCLE(770,-803,17100))" // &
            "||((DETX,DETY) in BOX(-20,-17000,6500,500,0))" // &
            "||((DETX,DETY) in BOX(5880,-20500,7500,1500,10))" // &
            "||((DETX,DETY) in BOX(-5920,-20500,7500,1500,350))" // &
            "||((DETX,DETY) in BOX(-20,-20000,5500,500,0)))"
       chipdef=(/ "((DETX,DETY) in BOX(20,-60,6610,6570,0))          ",&
                  "((DETX,DETY) in BOX(6550,-13572,6590,6599,0))     ",&
                  "((DETX,DETY) in BOX(13280,-306,6610,6599,0))      ",&
                  "((DETX,DETY) in BOX(6700,13070,6530,6560,0))      ",&
                  "((DETX,DETY) in BOX(-6410,13130,6570,6560,0))     ",&
                  "((DETX,DETY) in BOX(-13169,-105,6599,6599,0))     ",&
                  "((DETX,DETY) in BOX(-6540,-13438,6570,6599,0))    " /)
       fulldef = "&&(((DETX,DETY) IN box(-2683.5,-15917,2780.5,1340,0))" // &
            "||((DETX,DETY) IN box(2743.5,-16051,2579.5,1340,0))" // &
            "||((DETX,DETY) IN circle(97,-172,17152)))"
    else if (inst .eq. 'EMOS2') then ! EMOS2
       write (errstr,*) "Defining corner selection expression for EMOS2"
       call message(AppMsg,Noisy)
       cornerdef = "&&!(((DETX,DETY) IN CIRCLE(435,1006,17100))" // &
            "||((DETX,DETY) IN CIRCLE(-34,68,17700))" // &
            "||((DETX,DETY) IN BOX(-20,-17000,6500,500,0))" // &
            "||((DETX,DETY) IN BOX(5880,-20500,7500,1500,10))" // &
            "||((DETX,DETY) IN BOX(-5920,-20500,7500,1500,350))" // &
            "||((DETX,DETY) IN BOX(-20,-20000,5500,500,0)))"
       chipdef = (/ "((DETX,DETY) IN BOX(20,-60,6610,6590,0))        ", &
            "((DETX,DETY) IN BOX(6580,-13530,6620,6640,0))   ", &
            "((DETX,DETY) IN BOX(13320,-295,6620,6590,0))    ", &
            "((DETX,DETY) IN BOX(6610,13110,6590,6550,0))    ", &
            "((DETX,DETY) IN BOX(-6560,13180,6590,6600,0))   ", & 
            "((DETX,DETY) IN BOX(-13190,-90,6600,6630,0))    ", &
            "((DETX,DETY) IN BOX(-6620,-13438,6620,6599,0))  " /);
       fulldef="&&(((DETX,DETY) IN circle(-61,-228,17085))" // &
            "||((DETX,DETY) IN box(14.375,-16567.6,5552.62,795.625,0)))"
    else
       call error("notM1M2PN","Instrument is not M1, M2, or even PN, weirdly")
    end if
    
    return

  end subroutine defineexpr

  subroutine read_region(regfile,region,withregion)
    !
    ! read the text region file to create fovdef
    ! Get the definition of the object region
    !
    character(len=*) :: regfile, region
    logical :: i_ok = .FALSE.
    logical, intent(inout) :: withregion
    integer :: ios, lun
    region = ""
    ios = 0
    inquire(FILE=trim(adjustl(regfile)), EXIST=i_ok)
    if (i_ok) then
       write(errstr,*) "Region file exists: ", trim(regfile)
       call message(AppMsg,Noisy)
       lun = GetLun() ! in gofutils
       open (lun,file=trim(regfile), iostat=ios)
       if (ios .ne. 0) then
          write(errstr,*)"Cannot OPEN region file, iostat= ",ios
          call error ("noOPEN",trim(adjustl(errstr)))
       end if
       read(lun,'(A)', iostat=ios)region ! do some sanity checking???!!!
       if (ios .ne. 0) then
          write(errstr,*)"Cannot READ region file, iostat= ",ios
          call error ("noREAD",trim(adjustl(errstr)))
       end if
       write(errstr,*) "Area selection from region text file: ",&
            trim(adjustl(region))
       call message(AppMsg,Always)
    else
       write(errstr,*)"Chosen region file: ",trim(adjustl(regfile))," does not exist, exiting..."
       call error("NoREGIONfile",trim(adjustl(errstr)))
    end if

    return

  end subroutine read_region
  !
  !
  !
  subroutine docorner(inE_F,inC_F,ccddef,corndef,pattern)

    character(len=*) :: inE_F, inC_F, ccddef, corndef, pattern
    character(len=4096) :: command
    logical :: die
    !
    die = .FALSE.
    call check_clobber(inC_F,die)
    ! check this is exact same as espfilt!!!???
    command = "evselect table=" // trim(adjustl(inE_F)) // ":EVENTS " //&
         "withfilteredset=yes expression='(PATTERN<=" // trim(adjustl(pattern)) // &
         trim(adjustl(ccddef)) // "&&((FLAG & 0x766a0f63) == 0)" // trim(adjustl(corndef)) // &
         "' filtertype=expression keepfilteroutput=yes updateexposure=yes " // &
         "filterexposure=yes filteredset=" // trim(adjustl(inC_F))
!    write(errstr,*) "Corner-only event list evselect command = "
!    call message(AppMsg,Always)
!    write(errstr,*) "Step 7 ",trim(adjustl(command))
!    call message(AppMsg,Always)
    call system(trim(command))
    return
    
  end subroutine docorner
  !
  !
  !
  subroutine makeImFile(evlist,imfile,maskit,inst,ccddefa,pattern)
    !
    ! Make the clean FOV image files in sky coords
    ! OLDNEW: mos1S001-obj-im.fits mos1S001-fovim.fits (same as cheese call???!!!)
    !
    character(len=*), intent(in) :: evlist, imfile, maskit, ccddefa, pattern
    character(len=2) :: inst
    character(len=4096) :: command
    integer :: nMOS1, nMOS2, nPN, elow, ehigh
    logical :: i_ok,die
    command = ""
    
    if ((inst .eq. 'M1') .or. (inst .eq. 'M2')) then 
       write(errstr,*) "Making FOV image files for " // inst
       call message(AppMsg,Always)
       die = .FALSE.
       call check_clobber(imfile,die)
       ! check this ???!!! Real32 vs Int32???
       ! emask does not like Int32, wants Real32, what about downstream???!!!
       ! ** emask: error (accessingAsWrongDataType), Accessing data mos1S001-fovimt.fits.PRIMARY
       ! of type Int32 as real32 is not supported
       ! OLDNEW: mos1S001-obj-im.fits mos1S001-fovimt.fits
       command = "evselect table=" // trim(adjustl(evlist)) // &
            ":EVENTS withfilteredset=yes expression=" // & 
            "'(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
            trim(adjustl(ccddefa)) // &
            trim(adjustl(maskit)) // "&&(PI in [300:8000])' filtertype=expression " // &
            "imagebinning='imageSize' imagedatatype='Int32' imageset=" // &
            trim(adjustl(imfile)) // " squarepixels=yes ignorelegallimits=yes " // &
            "withxranges=yes withyranges=yes xcolumn='X' ximagesize=900 " // &
            "ximagemax=48400 ximagemin=3401 ycolumn='Y' yimagesize=900 " // &
            "yimagemax=48400 yimagemin=3401 updateexposure=yes filterexposure=yes"
!       write(errstr,*) inst // " Step 8. FOV image evselect command = "
!       call message(AppMsg,Always) ! change to Verbose???!!!
       call system(trim(command))
    else ! should never happen, but...
       call error("notM1M2PN","Instrument is not M1, M2, or even PN, weirdly")
    end if

    return

  end subroutine makeImFile
  !
  !
  !
  subroutine makeSPfile(evlist,infile,maskit,fulldef,ccddef,fovdef,&
       celow,cehigh,pattern)
    !
    ! Extract the entire FOV region from observation data into an image
    ! Do it again but use all events.  This will be a mask for calculating
    ! the SP scaling
    ! Step 9.
    ! OLDNEW: mos1S001-obj-im-sp-det.fits mos1S001-fovimspdet.fits
    character(len=*), intent(in) :: evlist, infile, maskit, &
         ccddef, fulldef, fovdef, celow, cehigh, pattern
    character(len=4096) :: command
    logical :: die
    !
    write(errstr,*) "Making SP DET image file..."
    call message(AppMsg,Always) ! change to Noisy???!!!
    die = .FALSE.
    call check_clobber(infile,die)  ! goflib/gofio routine
    command = "evselect table=" // trim(adjustl(evlist)) // &
         ":EVENTS withfilteredset=yes expression=" // & 
         "'(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // &
         trim(adjustl(fulldef)) // trim(adjustl(ccddef)) // &
         trim(adjustl(maskit)) // trim(adjustl(fovdef)) // "' " // &
         "filtertype=expression imagebinning='imageSize' " // &
         "imagedatatype='Int32' imageset=" // trim(adjustl(infile)) // &
         " squarepixels=yes ignorelegallimits=yes withxranges=yes " // &
         "withyranges=yes xcolumn='DETX' ximagesize=780 ximagemax=19500 " // &
         "ximagemin=-19499 ycolumn='DETY' yimagesize=780 yimagemax=19500 " // &
         "yimagemin=-19499 updateexposure=yes filterexposure=yes"

    call system(trim(command))
    return

  end subroutine makeSPfile
  !
  subroutine makeBandImageDet (evlist,infile,maskit,fulldef,ccddef,ccddefa,fovdef,&
       celow,cehigh,pattern)
    ! Step 10.
    ! Extract the entire FOV region and band from observation data into an 
    ! image in detector coords (old: mos1S001-obj-im-det-350-1100.fits
    ! file will be mos1S001-fovim-350-1100-det.fits
    ! OLDNEW: mos1S001-obj-im-det-350-1100.fits mos1S001-fovimdet-350-1100.fits
    !
    character(len=*), intent(in) :: evlist, infile, maskit, pattern
    character(len=*), intent(inout) :: ccddef, ccddefa, fulldef, fovdef, celow, cehigh
    character(len=4096) :: command
    logical :: die
    !
    write(errstr,*) "Making FOV band image file..."
    call message(AppMsg,Always)
    call check_clobber(infile,die)  ! goflib/gofio routine
    command = "evselect table="// trim(adjustl(evlist)) // &
         ":EVENTS withfilteredset=yes filtertype=expression expression=" // &
         "'(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // trim(adjustl(ccddef)) // &
         trim(adjustl(maskit)) // "&&(PI in [" // trim(adjustl(celow)) // &
         ":" // trim(adjustl(cehigh)) // "])" // trim(adjustl(fulldef)) // &
         trim(adjustl(fovdef)) // "' imagebinning='imageSize' " // &
         "imagedatatype='Int32' imageset=" // trim(adjustl(infile)) // &
         " squarepixels=yes ignorelegallimits=yes withxranges=yes " // &
         "withyranges=yes xcolumn='DETX' ximagesize=780 ximagemax=19500 " // &
         "ximagemin=-19499 ycolumn='DETY' yimagesize=780 yimagemax=19500 " // &
         "yimagemin=-19499 updateexposure=yes filterexposure=yes"

!    write(errstr,*) " FOV region and band image evselect command = "
!    call message(AppMsg,Always)
!    write(errstr,*)"Step 10 ",trim(adjustl(command))
!    call message(AppMsg,Always)
    call system(trim(command))

    return

  end subroutine makeBandImageDet
  !
  subroutine makeBandImageSky (evlist,infile,maskit,fulldef,ccddef,ccddefa,fovdef,&
       celow,cehigh,pattern)
    !
    ! Extract the entire FOV region and band from observation data into an 
    ! image in sky coords (old: mos1S001-obj-im-det-350-1100.fits
    ! file will be mos1S001-fovim-350-1100-det.fits
    ! OLDNEW: mos1S001-obj-im-det-350-1100.fits mos1S001-fovim-350-1100-det.fits
    !
    character(len=*), intent(in) :: evlist, infile, maskit, pattern
    character(len=*), intent(inout) :: ccddef, ccddefa, fulldef, fovdef, celow, cehigh
    character(len=4096) :: command
    logical :: die

    write(errstr,*) "Making band image file in SKY coords..."
    call message(AppMsg,Always)
    die = .FALSE.
    call check_clobber(infile,die)  ! goflib/gofio routine
    command = "evselect table="// trim(adjustl(evlist)) // &
         ":EVENTS withfilteredset=yes filtertype=expression expression=" // &
         "'(PATTERN<=" // trim(adjustl(pattern)) // "&&(FLAG == 0)" // trim(adjustl(ccddefa)) // &
         trim(adjustl(maskit)) // "&&(PI in [" // trim(adjustl(celow)) // &
         ":" // trim(adjustl(cehigh)) // "])" // trim(adjustl(fulldef)) // &
         trim(adjustl(fovdef)) // "' imagebinning='imageSize' " // &
         "imagedatatype='Int32' imageset=" // trim(adjustl(infile)) // &
         " squarepixels=yes ignorelegallimits=yes withxranges=yes " // &
         "withyranges=yes xcolumn='X' ximagesize=900 ximagemax=48400 " // &
         "ximagemin=3401 ycolumn='Y' yimagesize=900 yimagemax=48400 " // &
         "yimagemin=3401 updateexposure=yes filterexposure=yes"
    !
    !write(errstr,*) "Step 16.a Sky coord Band image evselect command = "
    !call message(AppMsg,Always) ! change to Verbose???!!!
    !write(errstr,*) trim(adjustl(command))
    !call message(AppMsg,Always)
    call system(trim(command))

    return

  end subroutine makeBandImageSky

  subroutine gonogo(file,go)

    character(len=*) :: file
    logical, intent(inout) :: go 
    logical :: i_ok
    i_ok = .FALSE.
    go = .TRUE.
    inquire(FILE=trim(adjustl(file)), EXIST=i_ok)
    if (i_ok) then
       if ( clobber() ) then
          write(errstr,*) " File: ",trim(file),&
               " exists and CLOBBER is set, will overwrite existing file."
          call message(AppMsg,Always)          
          call unlink (trim(file))
       else
          write(errstr,*)  " File: ",trim(file),&
               " exists and CLOBBER is NOT set, using existing..."
          call message(AppMsg,Always)
          go = .FALSE.
       endif
    else 
       write(errstr,*)  " File: ",trim(file),&
            " does not exist, will create."
       call message(AppMsg,Always)       
    end if
    
    return

  end subroutine gonogo
  !
  !
  !
  subroutine getkeys(in_S,nRows,expo,livetime,schip)

    type(DataSetT) :: in_S
    type(TableT) :: in_T
    character(len=1), intent(in) :: schip
    character(len=8) :: livetimN
    integer, intent(out) :: nRows
    integer :: nBlocks, i
    logical :: gotit
    real(kind=single), intent(out) :: expo, livetime ! should be double, maybe???
    !
    livetimN = "LIVETI0" // schip
    livetime = 0.
    nRows = 0
    expo = 0.
    gotit = .FALSE.
    !
    if (.not.hasBlock(in_S,'EVENTS')) then
       call error("noEVENTS","Input file does not contain an EVENTS extension.")
    else
       in_T = table(in_S,"EVENTS")
       nRows = numberOfRows(in_T)
    endif
    call release(in_T)
    !
    nBlocks = numberOfBlocks(in_S)
    gotit = .FALSE.
    do i=0,nBlocks-1
       in_T = table(in_S,i)
       if (hasAttribute(in_T,'EXPOSURE')) then
          expo = real32attribute(in_T,'EXPOSURE')
          gotit = .TRUE.
       end if
       call release(in_T)
    enddo
    if (.not. gotit) then
       call warning("noEXPOSURE","Input file does not contain an EXPOSURE attribute")
    endif
    !
    in_T = table(in_S,"EVENTS")
    if (.not. hasAttribute(in_T,trim(adjustl(livetimN)))) then
       if (hasAttribute(in_S,trim(adjustl(livetimN)))) then
          livetime = real32attribute(in_S,trim(adjustl(livetimN)))
       else
          write(errstr,*)"no" // trim(adjustl(livetimN))
          call warning(trim(adjustl(errstr)),&
               "Input file does not contain a LIVETInn attribute")
       end if
    else
       livetime = real32attribute(in_T,trim(adjustl(livetimN)))
    endif
    call release(in_T)
    
    return

  end subroutine getkeys
  !
  !
  !
  subroutine anom(inst,chip,rate,hard,hardsig,ccds,newccds)

    character(len=*) :: inst
    character(len=1) :: schip
    integer :: newccds(7), chip
    logical, intent(inout) :: ccds (7)
    real :: rate, hard, temphard, hardsig
    write(schip,'(I1)') chip
    temphard = 0.
    newccds = 0
    if (inst .eq. 'M1') then
       if (chip .eq. 4) then
          if (rate .lt. 5.) then
             temphard = 0.90
          else
             if (rate .lt. 7.) then
                temphard = 0.25 * rate - 0.35
             else
                temphard = 1.4
             endif
          endif
          if ((hard+hardsig) .lt. temphard) then
             write(errstr,*) "MOS1 CCD #4 is in an anomolous state"
             call message(AppMsg,Always)
             write(errstr,*) "Here there be Dragons - Untested Analysis"
             call message(AppMsg,Always)
             newccds(4) = 0
          else
             if ((hard-hardsig) .lt. temphard) then
                write(errstr,*) "MOS1 CCD #4 is one sigma within the definition of an anomolous state"
                call message(AppMsg,Always)
                write(errstr,*) "Here there be Dragons"
                call message(AppMsg,Always)
                newccds(4) = 2
             endif
          end if
       else   
          if (chip .eq. 5) then
             temphard = 1.44 + 0.074 * rate
             if ((hard+hardsig) .lt. temphard) then
                write(errstr,*) "MOS1 CCD #5 is in an anomolous state"
                call message(AppMsg,Always)
                write(errstr,*) "Here there be Dragons - Untested Analysis"
                call message(AppMsg,Always)
                newccds(5) = 0
             else
                 if((hard-hardsig) .lt. temphard) then
                    write(errstr,*) "MOS1 CCD #5 is one sigma within the definition of an anomolous state"
                    call message(AppMsg,Always)
                    write(errstr,*) "Here there could be Dragons"
                    call message(AppMsg,Always)
                    newccds(5) = 2
                 end if
              endif
           else
               if((hard+hardsig) .lt. 1.0) then
                  write(errstr,*) "MOS1 CCD #" // schip // " is in an anomolous state"
                  call message(AppMsg,Always)
                  write(errstr,*) "Here there could be Dragons"
                  call message(AppMsg,Always)
                  newccds(chip) = 0
               else
                  if(hard-hardsig .lt. 1.0) then
                     write(errstr,*) "MOS1 CCD #" // schip // " is one sigma within the definition " // &
                          "of an anomolous state"
                     call message(AppMsg,Always)
                     write(errstr,*) "Here there be Dragons"
                     call message(AppMsg,Always)
                     newccds(chip) = 2
                  endif
               endif
            endif
         endif
      elseif (inst .eq. 'M2') then
       if (chip .eq. 2) then
          if((hard+hardsig) .lt. 1.0) then
             write(errstr,*) "MOS1 CCD #2 is in an anomolous state"
             call message(AppMsg,Always)
             write(errstr,*) "Here there could be Dragons"
             call message(AppMsg,Always)
             newccds(2) = 0
          else
              if(hard-hardsig .lt. 1.0) then
                 write(errstr,*) "MOS2 CCD #2 is one sigma within the definition " // &
                      "of an anomolous state"
                 call message(AppMsg,Always)
                 write(errstr,*) "Here there could be Dragons"
                 call message(AppMsg,Always)
                 newccds(2) = 2
              endif
           endif
        else
           if(chip .eq. 5) then
              if(rate .lt. 2.75)then
                 temphard=0.0
              else
                 if(rate .lt. 7.2) then
                    temphard = 0.18 * rate + 0.4056
                 else 
                    temphard=1.7
                 end if
              end if
              if((hard+hardsig) .lt. temphard) then
                 write(errstr,*) "MOS2 CCD #5 is in an anomolous state"
                 call message(AppMsg,Always)
                 write(errstr,*) "Here there could be Dragons"
                 call message(AppMsg,Always)
                 newccds(5) = 0
              else
                  if(hard-hardsig .lt. temphard) then
                     write(errstr,*) "MOS1 CCD #5 is one sigma within the definition " // &
                          "of an anomolous state"
                     call message(AppMsg,Always)
                     write(errstr,*) "Here there be Dragons"
                     call message(AppMsg,Always)
                     newccds(5) = 2
                  endif
               end if
            else
               if((hard+hardsig) .lt. 1.0) then
                  write(errstr,*) "MOS2 CCD #" // schip // " may be in an anomolous state"
                  call message(AppMsg,Always)
                  write(errstr,*) "Here there could be Dragons -- Untested Analysis"
                  call message(AppMsg,Always)
                  newccds(chip) = 0
               else
                   if(hard-hardsig .ge. 1.0) then
                      write(errstr,*) "MOS1 CCD #" // schip // " is one sigma within the definition " // &
                           "of an anomolous state"
                      call message(AppMsg,Always)
                      write(errstr,*) "Here there could be Dragons"
                      call message(AppMsg,Always)
                      newccds(5) = 2
                   endif
                endif
             endif
          endif
       endif
      
      return
 
    end subroutine anom
    !
    !
    !
    subroutine copyARFattributes(fov_S,ccd_S,expo)
      ! copy *fovt.pi to various fwc*.pi spectra to allow backscale/arfgen to function 
      type(DataSetT) :: fov_S,ccd_S
      character(len=25) :: dummString
      real :: expo
      real (kind=double) :: dummReal
      
      write(errstr,*)'copyARF: Writing FOV spectrum header keywords to FWC spectrum header'
      call message(AppMsg,Always) ! change to Verbose later???!!!
      dummString = stringAttribute(fov_S,"INSTRUME")
      call setAttribute(ccd_S,'INSTRUME',trim(adjustl(dummString)),"XMM mission")
      dummString = stringAttribute(fov_S,"TELESCOP")
      call setAttribute(ccd_S,'TELESCOP',trim(adjustl(dummString)),"EPIC MOS Instrument")
      dummString = stringAttribute(fov_S,"DATE-OBS")
      call setAttribute(ccd_S,'DATE-OBS',trim(adjustl(dummString)),"Start time of exposure")
      dummString = stringAttribute(fov_S,"DATE-END")
      call setAttribute(ccd_S,'DATE-END',trim(adjustl(dummString)),"End time of exposure")
      dummString = stringAttribute(fov_S,"DATAMODE")
      call setAttribute(ccd_S,'DATAMODE',trim(adjustl(dummString)),"Instrument mode (IMAGING or TIMING)")
      dummString = 'Closed'
      call setAttribute(ccd_S,'FILTER',trim(adjustl(dummString)),'Filter ID')
      !
      dummReal= real64Attribute(fov_S,"RA_PNT")
      call setAttribute(ccd_S,'RA_PNT',dummReal,"deg","Actual (mean) pointing RA of the optical")
      dummReal= real64Attribute(fov_S,"DEC_PNT")
      call setAttribute(ccd_S,'DEC_PNT',dummReal,"deg","Actual (mean) pointing Dec of the optical")
      call setAttribute(ccd_S,'EXPOSURE',expo,'sec','Exposure Time')
      
    end subroutine copyARFattributes
    
end module mosspectra_mod
